classdef FaultOscHandler<handle
    %FAULTOSCHANDLER Summary of this class goes here
    %   Detailed explanation goes here
    
    properties %(SetAccess=private, GetAccess=public)
        comtrade_data
        fourier_data
        
        N %   
        current_set_multiplier=3;
        
        %   
        num_prefault_sample 
        num_fault_sample
        num_afterfault_sample
        
        faultTypeResult
        
        % 
        fault_type
        %   
        U_norm
        I_norm
        
        U_fault
        I_fault
        
        faultContour
        
        time_limit=[];
    end
    
    methods
        
        function HandleOscillogramm(this,oscReadSettings, num_current_channels, num_Uchannels, num_Ichannels)
            this.comtrade_data=COMTRADE.ReadComtrade(oscReadSettings);
            this.SingalProcessFilter(num_Uchannels, num_Ichannels)
            this.FindFaultType(num_current_channels);
            this.GetFaultData(num_Uchannels, num_Ichannels);
        end
        
        function SingalProcessFilter(this, num_Uchannels,num_Ichannels) % -      
            f=this.comtrade_data.info.frequency;
            samp_rate=this.comtrade_data.info.samp_rate;
            this.N=samp_rate/f; %   
            w=2*pi*f;
            
            %  
            if ~isempty(this.time_limit)
                [first_sample, last_sample]=FaultOscHandler.HandleTimeLimits(this.comtrade_data.times, this.time_limit);
                samples=first_sample:last_sample;
            else
                samples=1:length(this.comtrade_data.times);
            end
            

            num_samples=length(samples);
            if num_samples<3*this.N
                error( ['   ,    ' num2str(3*this.N) '   '] );
            end
            last_ref_sample=num_samples-this.N+1;
            
            chosen_channels={};
            num_all_channels=[num_Uchannels num_Ichannels];
            for m=1:length(num_all_channels)                               
                    for i=1:length(this.comtrade_data.ana_channels)
                       if this.comtrade_data.ana_channels{i}.N==num_all_channels(m)
                          channel=this.comtrade_data.ana_channels{i};
                          channel.values=channel.values(samples);
                          channel.times=channel.times(samples);
                          chosen_channels{end+1}=channel;
                       end
                    end               
            end
            
            for i=1:length(chosen_channels)
                times=chosen_channels{i}.times;
                values=chosen_channels{i}.values;
                yDFT=zeros(1,last_ref_sample);
                for k=1:num_samples-this.N+1 
                    numbers=k:k+this.N-1;
                    t=times(numbers);
                    y=values(numbers);
                    yDFT(k)=sqrt(2)*1i/this.N*sum(y.*exp(-1i*w*t));
                end               
                this.fourier_data{i}.times=times(1:last_ref_sample);
                this.fourier_data{i}.values=yDFT;
                this.fourier_data{i}.N=chosen_channels{i}.N;%  !!!     N,      
                this.fourier_data{i}.ch_id=chosen_channels{i}.ch_id;%id 
            end
        end
        
        function FindFaultType(this, num_current_channels)%       
            %    
            %num_current_channels=[4 5 6] %
            % . 1,2,3 - , , 
            ch=this.GetFourierChannel(num_current_channels(1));            
            I{1}=ch.values;
            ch=this.GetFourierChannel(num_current_channels(2));            
            I{2}=ch.values;
            ch=this.GetFourierChannel(num_current_channels(3));            
            I{3}=ch.values;
            

            % " ",     
            for k=1:3
                I_prefault(k,1)=I{k}(1);
            end        
            a=exp(2i*pi/3);
            M_inv=1/3*...
                [1 1 1;
                1 a a^2
                1 a^2 a];
             M=M_inv^-1;
            
            Is_prefault=M_inv*I_prefault;            
            I0_prefault=Is_prefault(1);
            I1_prefault=Is_prefault(2);
            I2_prefault=Is_prefault(3);
            
            dead_zone=2;% ,     -    
            %         .
            %    .
            num_DFT_samples=length(I{1});

            %     
             for i=1: num_DFT_samples
                 Is=M_inv*[I{1}(i);I{2}(i);I{3}(i)];
                 I0(i)=Is(1);
                 I1(i)=Is(2);
                 I2(i)=Is(3);
             end
             %      0.1  I1max
             I_nb_set=0.1*max(abs(I1));
             I1_set=2*abs(I1_prefault);
             
             for i=1: num_DFT_samples
                 if abs(I0(i))<I_nb_set
                     I0_inc(i)=0;
                 else
                     I0_inc(i)=I0(i)-I0_prefault;
                 end
                 
                 if abs(I1(i))<I1_set||abs(I1(i)-I1_prefault)<I_nb_set
                     I1_inc(i)=0;
                 else
                     I1_inc(i)=I1(i)-I1_prefault;
                 end
                 
                 if abs(I2(i))<I_nb_set
                     I2_inc(i)=0;
                 else
                     I2_inc(i)=I2(i)-I2_prefault;
                 end
             end
             
             %  
             for i=1: num_DFT_samples
                    I_inc_filtered=M*[I0_inc(i);I1_inc(i);I2_inc(i)];
                    I_inc{1}(i)=I_inc_filtered(1);
                    I_inc{2}(i)=I_inc_filtered(2);
                    I_inc{3}(i)=I_inc_filtered(3);
             end

             % "" ,     
             filter_transient_status=zeros(1,num_DFT_samples);
             i=1;
             dN=dead_zone*this.N;
             num_last=length(I{1})-dead_zone*this.N;

             while true
                 if i>num_last
                     break
                 end
                 transient=false;
                 coeff=0;
                 for k=1:3
                     if I_inc{k}(i)==0&&I_inc{k}(i+1)~=0
                         transient=transient||true;
                     elseif I_inc{k}(i)~=0&&I_inc{k}(i+dN)==0
                         transient=transient||true;
                     end
                     %   
                     dy=abs(I_inc{k}(i))-abs(I_inc{k}(i+1));
%                      dt=time{k}(i)-time{k}(i+1);
%                      if abs(atan(dy/dt))*180/pi>=45; %      45
                    if dy>I_nb_set
                        transient=transient||true;
                    end
                 end
                     if I1_inc(i)==0&&I1_inc(i+1)~=0
                         transient=transient||true;
                     elseif I1_inc(i)~=0&&I1_inc(i+dN)==0
                         transient=transient||true;
                     end
                     %   
                     %                      dy=abs(I_inc{k}(i))-abs(I_inc{k}(i+1));
                     %                      dt=time{k}(i)-time{k}(i+1);
                     %                      if abs(atan(dy/dt))*180/pi>=45; %      45
                     %                          transient=transient||true;
                     %                      end
                 if transient
                     filter_transient_status(i:i+dN-1)=1;
                     i=i+dN;
                 else
                     i=i+1;
                 end
%                  corr_coeff(i)=coeff;
             end

             %    
             this.fault_type={};
             for i=1: num_DFT_samples
                 I_inc=M*[I0_inc(i); I1_inc(i); I2_inc(i)];
                 max_val=max( abs(I_inc));
                 min_val=0.1*max_val;
                 this.fault_type{i}='undef';
                 if filter_transient_status(i)==0
                     if abs(I0_inc(i))==0&&abs(I2_inc(i))==0&&abs(I1_inc(i))==0
                         this.fault_type{i}='none';
                     elseif   abs(I0_inc(i))==0&&abs(I2_inc(i))==0&&abs(I1_inc(i))~=0
                         this.fault_type{i}='ABC';
                     elseif abs(I0_inc(i))==0&&abs(I2_inc(i))~=0&&abs(I1_inc(i))~=0
                         % 
                         if  abs(I_inc(1))>min_val&&abs(I_inc(2))>min_val
                             this.fault_type{i}='AB';
                         elseif abs(I_inc(2))>min_val&&abs(I_inc(3))>min_val
                             this.fault_type{i}='BC';
                         elseif abs(I_inc(3))>min_val&&abs(I_inc(1))>min_val
                             this.fault_type{i}='CA';
                         end
                     elseif abs(I0_inc(i))~=0&&abs(I2_inc(i))~=0&&abs(I1_inc(i))~=0
                         single_phase=false;
                         dAB=abs(I_inc(1)-I_inc(2));
                         dBC=abs(I_inc(2)-I_inc(3));
                         dCA=abs(I_inc(3)-I_inc(1));
                         if dAB<min_val % Ia-Ib=0
                             this.fault_type{i}='C0';
                             single_phase=true;
                         elseif dBC<min_val % Ib-Ic=0
                             this.fault_type{i}='A0';
                             single_phase=true;
                         elseif dCA<min_val % Ic-Ia=0
                             this.fault_type{i}='B0';
                             single_phase=true;
                         end
                         if single_phase==false
                             if  (dAB>=dBC)&&(dAB>=dCA)
                                 this.fault_type{i}='AB0';
                             elseif (dBC>=dAB)&&(dBC>=dCA)
                                 this.fault_type{i}='BC0';
                             elseif (dCA>=dAB)&&(dCA>=dBC)
                                 this.fault_type{i}='CA0';
                             end
                         end
                     end
                 end
             end
             %   
             %%     
            this.faultTypeResult={};
            this.num_fault_sample=[];
            for i=1:num_DFT_samples
                if ~strcmp(this.fault_type{i},'none')&&~strcmp(this.fault_type{i},'undef')
                    this.faultTypeResult{end+1}=this.fault_type{i};
                    this.num_fault_sample(end+1)=i; %   ,   
                    %break
                end
            end
            
            this.num_prefault_sample=1;
            
             for i=1:(num_DFT_samples-2)
             no_transient=FaultOscHandler.ScanSamples(filter_transient_status, true, 2);  
                 if strcmp(this.fault_type{i},'none')&&(no_transient&&filter_transient_status(this.N+1)==1)
                     this.num_prefault_sample=i;
                     break
                 end
             end
             
             %    
             this.num_afterfault_sample=num_DFT_samples-2;
%              nums=(num_DFT_samples-2:-1:1);
             for i=num_DFT_samples:-1:3
                 
                 if strcmp(this.fault_type{i},'none')&&(filter_transient_status(i-1)==0&&filter_transient_status(i-2)==1)
                     this.num_afterfault_sample=i;
                     break
                 end
                 
             end
             
        end
        
        function GetFaultData(this, num_Uchannels, num_Ichannels)
            %num_Uchannels=[1 2 3] num_Ichannels=[4 5 6]   , 
            %-    -
            
            if length(num_Uchannels)~= length(num_Ichannels)
                error('     ')
            end
            if strcmp(this.faultTypeResult,'none')
                error('    ')
            end
            
            for i=1:length(num_Uchannels)
               n=num_Uchannels(i);
               channel=this.GetFourierChannel(n);
               this.U_norm(i,1)=channel.values(this.num_prefault_sample);
               for k=1:length(this.num_fault_sample)     
                    this.U_fault(i,k)=channel.values(this.num_fault_sample(k));
                end
            end
            for i=1:length(num_Ichannels)
                n=num_Ichannels(i);
                channel=this.GetFourierChannel(n);                
                this.I_norm(i,1)=channel.values(this.num_prefault_sample);
                for k=1:length(this.num_fault_sample)     
                    this.I_fault(i,k)=channel.values(this.num_fault_sample(k));
                end
            end
            
%             for k=1:length(this.faultTypeResult)
%                 if strcmp(this.faultTypeResult{k},'AB0')||strcmp(this.faultTypeResult{k},'ABC')
%                     this.faultTypeResult{k}='AB';
%                 elseif strcmp(this.faultTypeResult{k},'BC0')
%                     this.faultTypeResult{k}='BC';
%                 elseif strcmp(this.faultTypeResult{k},'CA0')
%                     this.faultTypeResult{k}='CA';
%                 end
%             end
        end

        function channel=GetFourierChannel(this, num)
            channel=[];
            for i=1:length(this.fourier_data)
                if this.fourier_data{i}.N==num
                    channel=this.fourier_data{i};
                    break
                end
            end
        end

        function [F, t]=FrequencyDetector(this, num_voltage_channels)
            if length(num_voltage_channels)~=3
                error(' 3  ');
            end
            %     
            n=this.num_afterfault_sample;
            Ua=DataLine( this.comtrade_data.ana_channels{num_voltage_channels(1)}.times(n:end), this.comtrade_data.ana_channels{num_voltage_channels(1)}.values(n:end) );
            Ub=DataLine( this.comtrade_data.ana_channels{num_voltage_channels(2)}.times(n:end), this.comtrade_data.ana_channels{num_voltage_channels(2)}.values(n:end) );
            Uc=DataLine( this.comtrade_data.ana_channels{num_voltage_channels(3)}.times(n:end), this.comtrade_data.ana_channels{num_voltage_channels(3)}.values(n:end) );
            
            U{1}=Ua-Ub;
            U{2}=Ub-Uc;
            U{3}=Uc-Ua;
            
            f1=this.comtrade_data.info.frequency;% 
            epsilon=1e-6;
            for i=1:3
                [F{i}, t{i}]=Frequency.FrequencyDetector(U{i}, this.N, f1, epsilon);
            end
        end
        
    end
    
    methods (Static)
        function A=DC_SuppressFilter3samp(t_p2s, y_p2s, N, w) 
            %t_p3s - , y_p3s- , N -    
            %     2 . ,    -
            t1=t_p2s(1:N-1);
            t2=t_p2s(2:N);
            t3=t_p2s(3:N+1);
            
            y1=y_p2s(1:N-1);
            y2=y_p2s(2:N);
            y3=y_p2s(3:N+1);
            
            x1=FaultOscHandler.FourierFilter(t1, y1, N, w);
            x2=FaultOscHandler.FourierFilter(t2, y2, N, w);
            x3=FaultOscHandler.FourierFilter(t3, y3, N, w);
            
            is_calculated=false;
            if x2-x1~=0
                k=(x3-x2)/(x2-x1);
                if 1-k~=0
                    A=(x2-k*x1)/(1-k);
                    is_calculated=true;
                end
            end
            if ~is_calculated||abs(A)/max(abs([x1 x2 x3]))>=10
                A=x1;
            end

        end
        
        function yDFT=FourierFilter(t, y, N, w)
            %t - , y- , N -    
            yDFT=2i/N*sum(y.*exp(-1i*w*t));
        end
        
        function [first_sample,last_sample]=HandleTimeLimits(times, time_limit)
            if length(time_limit)==1&&time_limit(1)>0
                lower_time_limit=0;
                upper_time_limit=time_limit(1);
            elseif length(time_limit)==2
                lower_time_limit=time_limit(1);
                upper_time_limit=time_limit(2);     
            else
                lower_time_limit=0;
                upper_time_limit=times(end);
            end
            
            if max(times)<lower_time_limit||min(times)>upper_time_limit
                error('   ')
            end
            
            num_samples=length(times);
            last_sample=num_samples;
            for i=1:num_samples
                if times(i)>=upper_time_limit
                    last_sample=i;
                    break
                end
            end
            first_sample=1;
            
            for i=1:num_samples
                if times(i)>=lower_time_limit
                    first_sample=i;
                    break
                end
            end
        end
        
        function bool_result=ScanSamples(sample_array, error_value, num_values) %   
            
            if num_values<length(sample_array)
                bool_result=false;
            else
                bool_result=true;
                for i=1:num_values
                    if  sample_array(i)==error_value
                        bool_result=false;
                        break
                    end
                end
            end
            
        end
        
        
    end
    
end

